%======================================================================================================
% plottrajektorien_nichtlinear - Plot der Trajektorien eines nichtlinearen Systems
%======================================================================================================
%   Die Funktion zeigt die Trajektorien eines nichtlinearen Systems mit zwei Zustandsgren 
%   an. Beide Zustandsgren sind abhngig von der Zeit, x1(t), x2(t) und werden bereinander 
%   dargestellt - x1(t) auf der x-Achse und x2(t) auf der y-Achse. 
%   
%   Betrachtet werden Systeme der Form:  vX'(t) = vF(vX, vU)  mit  vX = [x1, x2, ..., xn]
%
%   Berechnet wird der Verlauf von vorgegebenen Anfangswerten. Diese knnen entweder gleichmig
%   als Gitter oder auch nur ber den Rahmen eines bestimmten Bereiches verteilt sein. Berechnet 
%   werden die Zustandsgren im Bereich von tmin < t < tmax. Fr negative tmin werden Werte der 
%   Trajektorie vor dem Anfangspunkt zu berechnet. Dies ist sinnvoll, um Bereich instabiler 
%   Punkte besser auflsen zu knnen.
%
% SYNOPSIS
%  	plottrajektorien_nichtlinear( sDX1, sDX2, vBereich, vLaufZeit, bNeuesDiagramm, ... 
%                            bAnfangswertGitter, vAnfangswertBereich, vAnzahlAnfangswerte,...
%                            LaengePfeile )
%
% 	sDX1, sDX2          - Systemgleichungen (String): sDX1 = f1(x1, x2), sDX2 = f2(x1, x2)
%                         z.B.: x1'(t) = x2  ;  x2'(t) = -x1 + (1 - x1^2) * x2
%                            =>  sDX1 = 'x2'  ;  sDX2 = '-x1 + (1 - x1^2) * x2'                         
%   vBereich            - Vektor mit Darstellungsbereich: [xmin, xmax, ymin, ymax].
%   vLaufzeit           - Laufzeit der Trajektorien [tmin, tmax]. Fr negative tmin erfolgt
%                         also eine Rckwrtsberechnung der Trajektorie.
%   bNeuesDiagramm      - logical(1): neues Diagramm wird erstellt; logical(0): aktuelles Diagramm nutzen.
% 	bAnfangswertGitter  - logical(1): Gitter mit Anfangswerten erstellen; logical(0): nur Rahmen erstellen
%                         (ein Gitter ist z.B. sinnvoll bei elipsfrmigen Trajektorien).
%   vAnfangswertBereich - Vektor mit dem Bereich, in dem die Anfangswerte verteilt werden.
%   vAnzahlAnfangswerte - Anzahl der Anfangswerte in der Form [Nx, Ny]. Fr einen Rahmen betrgt 
%                         die Gesamtzahl der Anfangswerte dann N = 2 * (Nx + Ny), fr ein 
%                         Gitter N = Nx * Ny..
%   LaengePfeile        - Relative Lnge der Pfeile - 0 = keine Pfeile, 1 = normale Einstellung.
%
% BEISPIELE
%   PlotTrajektorien_Nichtlinear( 'x2', '-x1 + (1 - x1^2) * x2', [-3 3 -3 3], [-10 10], logical(1), ... 
%                                 logical(1), [1 1 1 1], [1 1], 1 )
%
%
%======================================================================================================
%  Autor:            Thomas Lens
%  Letzte nderung:  21.12.2004
%======================================================================================================
function PlotTrajektorien_Nichtlinear( sDX1_arg, sDX2_arg, vBereich_arg, vLaufzeit, bNeuesDiagramm, ... 
                                  bAnfangswertGitter, vAnfangswertBereich, vAnzahlAnfangswerte, ...
                                  LaengePfeile )

  %======================================================================================================
  % Globale Variablen deklarieren
  %======================================================================================================
  global sDX1;
  global sDX2;
  global vBereich;
  
  %======================================================================================================
  % Standartparameter nehmen, wenn keine Argumente bergeben wurden (erlaubt das Ausfhren aus dem Editor)
  %======================================================================================================
  if (nargin == 0)
    % --= Definition des nichtlinearen Systems ==- %
    sDX1 = 'x2';                                 % Nichtlineares System definieren.
    sDX2 = '-x1 + (1 - x1^2) * x2';
    % --= Alternatives Systems ==- %
    %sDX1 = 'x1 * (x2 - 1)';
    %sDX2 = 'x2 * (x1 - 1)';
    
    % --= Allgemeine Einstellungen =-- %
    vBereich = [-3, 3, -3, 3];                   % Darstellungsbereich in der Form [xmin, xmax, ymin, ymax].
    vLaufzeit = [-2, 5];                         % Laufzeit der Trajektorien der einzelnen Anfangswerte.
    bNeuesDiagramm = logical(1);                       % Wenn gesetzt, wird ein neues Diagramm erstellt. Sonst aktuelles Diagramm nutzen.
    
    % --= Einstellungen fr Verteilung der Anfangswerte =-- %
    bAnfangswertGitter = logical(1);                  % Wenn nicht gesetzt, wird nur ein Anfangswertrahmen erstellt.
    vAnfangswertBereich = [-3, 3, -3, 3];               
    vAnzahlAnfangswerte = [3, 3];               % Anzahl der Anfangswerte in der Form [Nx, Ny].
    
    % -== Einstellungen zu den Pfeilen ==-
    LaengePfeile = 1;                          % Relative Lnge der Pfeile.
        
  elseif (nargin ~= 9)
    % -== Inkorrekter Funktionsaufruf ==-
    error('Inkorrekte Anzahl an Argumenten.');   % Korrekte Anzahl der Parameter berprfen.
    return
    
  else
    % -== Bei korrektem Funktionsaufruf Argumente an die globalen Variablen bergeben ==-
    sDX1 = sDX1_arg;
    sDX2 = sDX2_arg;
    vBereich = vBereich_arg;
    
  end
 
  %======================================================================================================  
  % Argumente berprfen
  %======================================================================================================
  if ((vBereich(2) <= vBereich(1)) || (vBereich(4) <= vBereich(3)))
    error('Parameter vBereich = [xmin, xmax, ymin, ymax] hat fehlerhafte Werte.');
    return
  elseif ((vLaufzeit(1) > 0) || (vLaufzeit(2) < 0))
    error('Parameter vLaufzeit = [tmin, tmax] hat fehlerhafte Werte.')
    return
  elseif (LaengePfeile < 0)
    error('Parameter LaengePfeile ist fehlerhaft: 0 <= l < 10.');
    return
  elseif ((vAnfangswertBereich(2) < vAnfangswertBereich (1)) || (vAnfangswertBereich (4) < vAnfangswertBereich(3) > 0))
    error('Parameter vAnfangswertBereich = [xmin, xmax, ymin, ymax] hat fehlerhafte Werte.');
    return
  elseif ((vAnzahlAnfangswerte(1) < 0) || (vAnzahlAnfangswerte(2) < 0))
    error('Parameter vAnzahlAnfangswerte = [Nx, Ny] ist fehlerhaft, Nx und Ny drfen nicht negativ sein.'); 
    return
  end
  
  % -== Semikolon an Systemausdruck anhngen ==-
  if (sDX1(length(sDX1)) ~= ';'), sDX1 = [sDX1 ';']; end
  if (sDX2(length(sDX2)) ~= ';'), sDX2 = [sDX2 ';']; end
    
  % -== Systemausdruck testen ==-
  x1 = 1; x2 = 1;
  try, eval(sDX1); catch error('Parameter sDX1 ist fehlerhaft und kann nicht ausgewertet werden'); end
  try, eval(sDX2); catch error('Parameter sDX2 ist fehlerhaft und kann nicht ausgewertet werden'); end
  
  %======================================================================================================  
  % Initialisierung des Diagramms
  %======================================================================================================
  if (bNeuesDiagramm == logical(1)) 
    figure;                                      % Neues Diagramm erstellen, wenn dies erwnscht ist.
  end                             
  hold on;                                       % Alle Trajektorien werden im selben Diagramm ausgegeben.
  axis equal;
  axis(vBereich);                                % Achsen auf den gewnschten Bereich festlegen.
  title('Trajektorien fr nichtlineares System');    % Titel des Diagramms.
  xlabel('Zustandsgre x_{1}');                 % Achsenbeschriftung x-Achse.
  ylabel('Zustandsgre x_{2}');                 % Achsenbeschriftung y-Achse.
  
  %======================================================================================================  
  % Anfangswerte berechnen, verteilt ber Rahmen oder Gitter
  %======================================================================================================
  if (~bAnfangswertGitter)                % Anfangswertrahmen berechnen.
    mX0 = AnfangswertRahmen(vAnfangswertBereich, vAnzahlAnfangswerte);
  else                                          % Anfangswertgitter berechnen.
    mX0 = AnfangswertGitter(vAnfangswertBereich, vAnzahlAnfangswerte);
  end

  %======================================================================================================  
  % Berechnung und Ausgabe der Trajektorien und Richtungspfeile
  %======================================================================================================
  BerechneTrajektorien(sDX1, sDX2, vLaufzeit, LaengePfeile, mX0);
  
return




%======================================================================================================
% AnfangswertRahmen - Anfangswertpunkte auf dem Rand eines vorgegebenen Bereichs berechnen.
%======================================================================================================
%   Die Funktion berechnet eine [N x 2]-Matrix mit N Vektoren, die auf dem Rand eines
%   vorgegebenen Bereichs verteilt sind.
% 
% SYNOPSIS
%  	mX0 = AnfangswertRahmen( vBereich, vAnzahlAnfangswerte )
%                            
%   vBereich            - Bereich fr die zu verteilenden Anfangswerte.
%   vAnzahlAnfangswerte - Anzahl der Anfangswerte in der Form [Nx, Ny]. Fr einen Rahmen betrgt 
%                         die Gesamtzahl der Anfangswerte N = 2 * (Nx + Ny).
%   mX0                 - [N x 2]-Matrix mit den N berechneten Anfangsvektoren.
%
%======================================================================================================
function mX0 = AnfangswertRahmen( vBereich, vAnzahlAnfangswerte )
  
  mX0_oben(:, 1) = linspace(vBereich(1), vBereich(2), vAnzahlAnfangswerte(1))';  % Oben, x-Werte.
  mX0_oben(:, 2) = vBereich(4);                                                  % Oben, y-Werte.
 
  mX0_unten(:, 1) = mX0_oben(:, 1);                                              % Unten, x-Werte
  mX0_unten(:, 2) = vBereich(3);                                                 % Unten, y-Werte
  
  mX0_links(:, 2) = linspace(vBereich(3), vBereich(4), vAnzahlAnfangswerte(2))'; % Links, y-Werte.
  mX0_links(:, 1) = vBereich(1);                                                 % Links, x-Werte.
  
  mX0_rechts(:, 2) = mX0_links(:, 2);                                            % Rechts, y-Werte.
  mX0_rechts(:, 1) = vBereich(2);                                                % Rechts, x-Werte.
  
  mX0 = [mX0_oben; mX0_rechts; mX0_unten; mX0_links];  % Vektor mit allen Anfangswerten, ber Rahmen verteilt.
  
return




%======================================================================================================
% AnfangswertGitter - Anfangswertpunkte auf einem Gitter in einem bestimmten Bereich berechnen.
%======================================================================================================
%   Die Funktion berechnet eine [N x 2]-Matrix mit N Vektoren, die auf einem Gitter in einem
%   vorgegebenen Bereichs verteilt sind.
% 
% SYNOPSIS
%  	mX0 = AnfangswertGitter( vBereich, vAnzahlAnfangswerte )
%                            
%   vBereich            - Bereich fr die zu verteilenden Anfangswerte.
%   vAnzahlAnfangswerte - Anzahl der Anfangswerte in der Form [Nx, Ny]. Fr ein Gitter betrgt 
%                         die Gesamtzahl der Anfangsvektoren N = (Nx * Ny).
%   mX0                 - [N x 2]-Matrix mit den N berechneten Anfangsvektoren.
%
%======================================================================================================  
function mX0 = AnfangswertGitter( vGBereich, vAnzahlAnfangswerte )
 
  vN = linspace(vGBereich(1), vGBereich(2), vAnzahlAnfangswerte(1));  % Vektor mit x-Werten des Rasters.
  vM = linspace(vGBereich(3), vGBereich(4), vAnzahlAnfangswerte(2));  % Vektor mit y-Werten des Rasters.
    
  mX0_Zeile(vAnzahlAnfangswerte(1), 2) = 0;      % Speicher fr eine Zeile des Anfangswertvektors allokieren.
        
  mX0_Zeile(:, 1) = vN';                         % Reihen des Gitters vorbereiten: x-Werte der Spalten speichern.
  mX0_Zeile(:, 2) = vM(1);                       % y-Wert der ersten Reihe eintragen. => mX0_Zeile enhlt Gitterwerte fr erste Zeile.
    
  mX0 = mX0_Zeile;                               % Anfangswertvektor mit Werten der ersten Zeile initialisieren.
  for (i = 2 : vAnzahlAnfangswerte(2))
    mX0_Zeile(:, 2) = vM(i);                     % Aktuelle Zeile: y-Wert ndert sich.
    mX0 = [mX0; mX0_Zeile];                      % Aktuelle Zeile zum Anfangswertvektor hinzufgen.
  end
  
return




%======================================================================================================
% BerechneTrajektorien - Berechnung der Trajektorien des nichtlinearen Systems
%======================================================================================================
% Behandelte nichtlineare Systeme: vX' = f(vX, vU)
%
% Es ergibt sich also fr einen zweidimensionalen Zustandsvektor und mit vU = 0:
%   dx1 = f1(x1, x2) * dt
%   dx2 = f2(x1, x2) * dt
%
% SYNOPSIS
%  	BerechneTrajektorien( sDX1, sDX2, vLaufzeit,ZeitAufloesung, LaengePfeile, AbstandPfeile, mX0 )
%                            
%   sDX1, sDX2     - Ausdruck mit der Systembeschreibung.
%   vLaufzeit      - Laufzeit der Trajektorien.
%   LaengePfeile   - Relative Lnge der Pfeile.
%   mX0            - [N x 2]-Matrix mit den N Anfangsvektoren.
%
%======================================================================================================  
function BerechneTrajektorien( sDX1, sDX2, vLaufzeit, LaengePfeile, mX0 )
  
  % -== Globale Variablen sichtbar machen ==-
  global vBereich  

  %======================================================================================================  
  % Vorwrtsberechnung - Hauptschleife zur Berechnung der Trajektorien fr jeden Anfangswert
  %======================================================================================================  
  if (vLaufzeit(2) > 0)
    for (n = 1 : size(mX0, 1))
      vX0 = mX0(n, :);
      [vT, mX] = ode45(@DGLVorwaerts, [0, vLaufzeit(2)], vX0, odeset('MaxStep', 3));
      plot(mX(:,1), mX(:, 2));
      
      % -== Berechnung des Pfeils ==- %
      if (size(mX, 1) >= 5)
        % -== Pfeilspitze und Ende des Pfeils ==- %
        vX2 = mX(5, :);                           % Spitze des Pfeils        
        vX1 = mX(4, :);                           % Ende des Pfeils   
        vP = (vX2 - vX1);                         % Lnge neu berechnen
        if (any(vP))
          vP = vP * LaengePfeile  * (vBereich(2) - vBereich(1)) * 0.02 / norm(vP);
        end
        vX1 = vX2 - vP;
        % -== Koordinaten des Dreiecks berechnen ==- %
        vD(1) = -vP(2);
        vD(2) = vP(1);
        vD = vD * 0.2;
        vP = [vX1(1) + vD(1), vX1(2) + vD(2); vX2(1), vX2(2); vX1(1) - vD(1), vX1(2) - vD(2); vX1(1) + vD(1), vX1(2) + vD(2)];
        % -== Richtungspfeil plotten ==- %
        fill(vP(:, 1), vP(:, 2), 'b');    
      end
      
    end
  end
  
  %======================================================================================================  
  % Rckwrtsberechnung - Hauptschleife zur Berechnung der Trajektorien fr jeden Anfangswert
  %======================================================================================================  
  if (vLaufzeit(1) < 0)
    for (n = 1 : size(mX0, 1))
      vX0 = mX0(n, :);
      [vT, mX] = ode45(@DGLRueckwaerts, [vLaufzeit(1) 0], vX0);
      plot(mX(:,1), mX(:, 2));
      
      % -== Berechnung des Pfeils ==- %      % Fr Rckwrtsberechnung keine Pfeile berechnen
%       if (size(mX, 1) >= 5)
%         % -== Pfeilspitze und Ende des Pfeils ==- %
%         vX2 = mX(4, :);                           % Spitze des Pfeils
%         vX1 = mX(5, :);                           % Ende des Pfeils
%         vP = (vX2 - vX1) * LaengePfeile;          % Lnge neu berechnen
%         vX1 = vX2 - vP;
%         % -== Koordinaten des Dreiecks berechnen ==- %
%         vD(1) = -vP(2);
%         vD(2) = vP(1);
%         vD = vD * 0.2;
%         vP = [vX1(1) + vD(1), vX1(2) + vD(2); vX2(1), vX2(2); vX1(1) - vD(1), vX1(2) - vD(2); vX1(1) + vD(1), vX1(2) + vD(2)];
%         % -== Richtungspfeil plotten ==- %
%         fill(vP(:, 1), vP(:, 2), 'b');    
%       end   
      
    end
  end
  
return




%======================================================================================================
% DGLVorwaerts - Differentialgleichung bei der Vorwrtsberechnung
%======================================================================================================
% Die Funktion gibt den aktuellen Wert von vX' = f(vX, vU) zurck.
%
% SYNOPSIS
%  	vDXDT = DGLVorwaerts(t, vX)
%                            
%   t     - Aktueller Zeitpunkt t (wird zur Berechnung nicht bentigt, wichtig fr Verwendung von ode45)
%   vX    - Aktueller Zustandsvektor vX(t)
%   vDXDT - Berechneter Wert vX'(t)
%   
%======================================================================================================  
function vDXDT = DGLVorwaerts(t, vX)

  % -== Globale Variablen sichtbar machen ==-
  global sDX1;
  global sDX2;
  global vBereich;
  
  vTBereich = vBereich * 2;
  % -== Bereich kontrollieren - luft die Kurve aus dem sichtbaren Bereich, dann abbrechen ==-
  if ((vX(1) < vTBereich(1)) || (vX(1) > vTBereich(2)) || (vX(2) < vTBereich(1)) || (vX(2) > vTBereich(2)))
    vDXDT = [0; 0];
    return
  end

  % -== Berechnung von vX' ==-
  x1 = vX(1);
  x2 = vX(2);
  eval(['dx1dt = ' sDX1]);
  eval(['dx2dt = ' sDX2]);
  vDXDT = [dx1dt; dx2dt];

return



%======================================================================================================
% DGLRueckwaerts - Differentialgleichung bei der Rckwrtsberechnung
%======================================================================================================
% Die Funktion gibt den aktuellen Wert von -vX' = -f(vX, vU) zurck.
%
% SYNOPSIS
%  	vDXDT = DGLRueckwaerts(t, vX)
%                            
%   t     - Aktueller Zeitpunkt t (wird zur Berechnung nicht bentigt, wichtig fr Verwendung von ode45)
%   vX    - Aktueller Zustandsvektor vX(t)
%   vDXDT - Berechneter Wert -vX'(t)
%   
%======================================================================================================  
function vDXDT = DGLRueckwaerts(t, vX)
  
  % -== Globale Variablen sichtbar machen ==-
  global sDX1;
  global sDX2;
  global vBereich;
  
  vTBereich = vBereich * 2;
  % -== Bereich kontrollieren - luft die Kurve aus dem sichtbaren Bereich, dann abbrechen ==-
  if ((vX(1) < vTBereich(1)) || (vX(1) > vTBereich(2)) || (vX(2) < vTBereich(1)) || (vX(2) > vTBereich(2)))
    vDXDT = [0; 0];
    return
  end
  
  % -== Berechnung von vX' ==-
  x1 = vX(1);
  x2 = vX(2);
  eval(['dx1dt = ' sDX1]);
  eval(['dx2dt = ' sDX2]);
  vDXDT = [-dx1dt; -dx2dt];

return